                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*  version 7: iterative constructs  */
                     /*-----------------------------------*/
                     /*              compiler             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipMain.h"

#include "SlipCompile.h"

/*------------------------------ private macros ------------------------------*/

#define chunk_size(TYP)                                                        \
  ((sizeof(TYP) - sizeof(CNT))/ sizeof(BTS_type))

/*------------------------------ private constants ---------------------------*/

static const TXT_type AL1_error_string = "at least one expression required";
static const TXT_type E1X_error_string = "exactly one expression required";
static const TXT_type ILT_error_string = "illegal expression type";
static const TXT_type ITF_error_string = "improperly terminated form";
static const TXT_type IPA_error_string = "invalid parameter";
static const TXT_type IVP_error_string = "invalid pattern in form";
static const TXT_type IVV_error_string = "missing variable in form";
static const TXT_type IXT_error_string = "invalid expression type";
static const TXT_type MSP_error_string = "missing pattern in form";
static const TXT_type MSV_error_string = "invalid variable in form";
static const TXT_type TFX_error_string = "too few expressions in form";
static const TXT_type TMX_error_string = "too many expressions in form";

/*------------------------------ private prototypes --------------------------*/

static EXP_type compile_expression(EXP_type  );
static VEC_type compile_parameters(PAI_type  ,
                                   EXP_type *,
                                   UNS_type  );
static VEC_type   compile_sequence(PAI_type  ,
                                   UNS_type  );

/*----------------------------------------------------------------------------*/
/*------------------------------------ application ---------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_application(EXP_type Operator,
                                    EXP_type Operands)
  { APL_type compiled_application;
    EXP_type compiled_operator;
    TAG_type tag;
    VEC_type compiled_operands; 
    compiled_operator = compile_expression(Operator);
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          compiled_operands = Main_Empty_Vector;
          break; 
        case PAI_tag:
          compiled_operands = compile_sequence(Operands,
                                               1);
          break; 
        default:
          return Main_Error_Text(ITF_error_string, 
                                 Main_Application_String); }
    compiled_application = make_APL(compiled_operator,
                                    compiled_operands);
    return compiled_application; }

/*----------------------------------------------------------------------------*/
/*------------------------------------ begin ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_begin(PAI_type Operands)
  { BEG_type compiled_begin;
    TAG_type tag;
    VEC_type compiled_sequence;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(AL1_error_string,
                                 Main_Begin_String); 
        case PAI_tag:
          compiled_sequence = compile_sequence(Operands,
                                               1);
          compiled_begin = make_BEG(compiled_sequence);
          return compiled_begin; }
    return Main_Error_Text(ITF_error_string,
                           Main_Begin_String); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ define --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_define(PAI_type Operands)
  { EXP_type compiled_define,
             compiled_expression,
             expression,
             pattern,
             residue;
    PAI_type body,
             expressions,
             parameters;
    SYM_type parameter,
             vararg,
             variable;
    TAG_type tag;
    VEC_type compiled_body,
             compiled_parameters;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(MSP_error_string,
                                 Main_Define_String);
        case PAI_tag:
          pattern = Operands->car;
          body    = Operands->cdr;
          tag = Tag_of(body);
          switch (tag)
            { case NUL_tag:
                return Main_Error_Text(AL1_error_string,
                                       Main_Define_String);
              case PAI_tag:
                tag = Tag_of(pattern);
                switch (tag)
                  { case PAI_tag:
                      expressions = (PAI_type)pattern;
                      variable   = expressions->car;
                      if (is_SYM(variable))
                        { parameters = expressions->cdr;
                          compiled_body = compile_sequence(body,
                                                           1);
                          tag = Tag_of(parameters);
                          switch (tag)
                            { case NUL_tag:
                                compiled_define = make_DFF(variable, 
                                                           Main_Empty_Vector, 
                                                           compiled_body);
                                return compiled_define;
                              case PAI_tag:
                                compiled_parameters = compile_parameters(parameters,
                                                                         &residue,
                                                                         1);
                                if (is_SYM(residue))
                                  { parameter = (SYM_type)residue;
                                    compiled_define = make_DFZ(variable, 
                                                               compiled_parameters, 
                                                               parameter,
                                                               compiled_body);
                                    return compiled_define; }
                                compiled_define = make_DFF(variable, 
                                                           compiled_parameters, 
                                                           compiled_body);
                                return compiled_define; 
                              case SYM_tag:
                                vararg = (SYM_type)parameters;
                                compiled_define = make_DFZ(variable, 
                                                           Main_Empty_Vector, 
                                                           vararg,
                                                           compiled_body);
                                return compiled_define; }
                          return Main_Error_Text(IPA_error_string, 
                                                 Main_Define_String); }
                      return Main_Error_Text(IVV_error_string, 
                                             Main_Define_String);
                    case SYM_tag:
                      variable = (SYM_type)pattern;
                      expression = body->car;
                      residue    = body->cdr;
                      if (is_NUL(residue))
                        { compiled_expression = compile_expression(expression);
                          compiled_define = make_DFV(variable, 
                                                     compiled_expression);
                          return compiled_define; }
                      return Main_Error_Text(E1X_error_string, 
                                             Main_Define_String); }
                return Main_Error_Text(IVP_error_string,
                                       Main_Define_String); }
          return Main_Error_Text(ITF_error_string,
                                 Main_Define_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_Define_String); }
                           

/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_if(PAI_type Operands)
  { EXP_type alternative,
             compiled_alternative,
             compiled_consequent,
             compiled_if,
             compiled_predicate,
             consequent,
             predicate,
             residue;
    PAI_type alternatives,
             clauses;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(TFX_error_string,
                                 Main_If_String);
        case PAI_tag:
          predicate = Operands->car;
          clauses   = Operands->cdr;
          tag = Tag_of(clauses);
          switch (tag)
            { case NUL_tag:
                return Main_Error_Text(TFX_error_string,
                                       Main_If_String);
              case PAI_tag:
                consequent   = clauses->car;
                alternatives = clauses->cdr;
                compiled_predicate = compile_expression(predicate);
                compiled_consequent = compile_expression(consequent);
                tag = Tag_of(alternatives);
                switch (tag)
                  { case NUL_tag:
                      compiled_if = make_IFZ(compiled_predicate,
                                             compiled_consequent);
                      return compiled_if;                  
                    case PAI_tag:
                      alternative = alternatives->car;
                      residue     = alternatives->cdr;
                      if (is_NUL(residue))
                        { compiled_alternative = compile_expression(alternative);
                          compiled_if = make_IFF(compiled_predicate,
                                                 compiled_consequent,
                                                 compiled_alternative);
                          return compiled_if; }
                      return Main_Error_Text(TMX_error_string,
                                             Main_If_String); }
                return Main_Error_Text(ITF_error_string,
                                       Main_If_String); }
          return Main_Error_Text(ITF_error_string,
                                 Main_If_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_If_String); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ lambda --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_lambda(PAI_type Operands)
  { EXP_type compiled_lambda,
             parameters,
             residue;
    PAI_type body;
    SYM_type parameter,
             vararg;
    TAG_type tag;
    VEC_type compiled_body,
             compiled_parameters;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(TFX_error_string,
                                 Main_Lambda_String);
        case PAI_tag:
          parameters = Operands->car;
          body       = Operands->cdr;
          compiled_body = compile_sequence(body,
                                           1);          
          tag = Tag_of(parameters);
          switch (tag)
            { case NUL_tag:
                compiled_lambda = make_LMB(Main_Empty_Vector, 
                                           compiled_body);
                return compiled_lambda;
              case PAI_tag:
                compiled_parameters = compile_parameters(parameters,
                                                         &residue,
                                                         1);
                if (is_SYM(residue))
                  { parameter = (SYM_type)residue;
                    compiled_lambda = make_LMZ(compiled_parameters, 
                                               parameter,
                                               compiled_body);

                    return compiled_lambda; }
                compiled_lambda = make_LMB(compiled_parameters, 
                                           compiled_body);
                return compiled_lambda; 
              case SYM_tag:
                vararg = (SYM_type)parameters;
                compiled_lambda = make_LMZ(Main_Empty_Vector, 
                                           vararg,
                                           compiled_body);
                return compiled_lambda; }
          return Main_Error_Text(IPA_error_string, 
                                 Main_Define_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_Lambda_String); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- parameters --------------------------------*/
/*----------------------------------------------------------------------------*/

static VEC_type compile_parameters(PAI_type   Parameters,
                                   EXP_type * Residue,
                                   UNS_type   Size)
  { EXP_type parameter;
    PAI_type parameters;
    TAG_type tag;
    VEC_type compiled_parameters;
    parameter  = Parameters->car;
    parameters = Parameters->cdr;
    if (is_SYM(parameter))
      { tag = Tag_of(parameters);
        switch (tag)
          { case PAI_tag:
              compiled_parameters = compile_parameters(parameters, 
                                                       Residue,
                                                       Size + 1);
              break;
            case NUL_tag:
              compiled_parameters = make_VEC(Size);
              *Residue = Main_Null;
              break;
            case SYM_tag:
              compiled_parameters = make_VEC(Size);
              *Residue = parameters;
              break;
            default:
              return Main_Error_Text(IPA_error_string,
                                     Main_Begin_String); }
        compiled_parameters[Size] = parameter; 
        return compiled_parameters; }
    return Main_Error_Text(IPA_error_string,
                           Main_Define_String); }
                                
/*----------------------------------------------------------------------------*/
/*------------------------------------ quote ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_quote(PAI_type Operands)
  { EXP_type expression,
             residue;
    QUO_type compiled_quote;
    expression = Operands->car;
    residue    = Operands->cdr;
    if (is_NUL(residue))
      { compiled_quote = make_QUO(expression);
        return compiled_quote; }                          /* clone!!! */
    return Main_Error_Text(TMX_error_string,
                           Main_Quote_String); }
                           
/*----------------------------------------------------------------------------*/
/*---------------------------------- sequence --------------------------------*/
/*----------------------------------------------------------------------------*/

static VEC_type compile_sequence(PAI_type Expressions,
                                 UNS_type Size)
  { EXP_type compiled_expression,
             expression;
    PAI_type expressions;
    TAG_type tag;
    VEC_type compiled_sequence;
    expression  = Expressions->car;
    expressions = Expressions->cdr;
    compiled_expression = compile_expression(expression);
    tag = Tag_of(expressions);
    switch (tag)
      { case NUL_tag:
          compiled_sequence = make_VEC(Size);
          break;
        case PAI_tag:
          compiled_sequence = compile_sequence(expressions, 
                                               Size + 1); 
          break;
        default:
          return Main_Error_Text(ITF_error_string,
                                 Main_Sequence_String); }
    compiled_sequence[Size] = compiled_expression; 
    return compiled_sequence; }

/*----------------------------------------------------------------------------*/
/*------------------------------------ set -----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_set(PAI_type Operands)
  { EXP_type compiled_expression,
             expression;
    PAI_type expressions,
             residue;
    SET_type compiled_set;
    SYM_type variable;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(MSV_error_string,
                                 Main_Set_String);
        case PAI_tag:
          variable = Operands->car;
          if (is_SYM(variable))
            { expressions = Operands->cdr;
              tag = Tag_of(expressions);
              switch (tag)
                { case NUL_tag:
                    return Main_Error_Text(E1X_error_string,
                                           Main_Set_String);
                  case PAI_tag:
                    expression = expressions->car;
                    residue    = expressions->cdr;
                    if (is_NUL(residue))
                      { compiled_expression = compile_expression(expression);
                        compiled_set = make_SET(variable,
                                                compiled_expression);
                        return compiled_set; }
                    return Main_Error_Text(TMX_error_string,
                                           Main_Set_String); }
              return Main_Error_Text(ITF_error_string,
                                     Main_Set_String); }
          return Main_Error_Text(IVV_error_string,
                                 Main_Set_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_Set_String); }
                           
/*----------------------------------------------------------------------------*/
/*------------------------------------ symbol --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_symbol(SYM_type Variable)
  { return Variable; }

/*----------------------------------------------------------------------------*/
/*----------------------------------- value ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_value(EXP_type Value)
  { return Value; }

/*----------------------------------------------------------------------------*/
/*----------------------------------- while ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_while(PAI_type Operands)
  { EXP_type body,
             compiled_body,
             compiled_predicate,
             predicate;
    TAG_type tag;
    WHI_type compiled_while;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(TFX_error_string,
                                 Main_While_String);
        case PAI_tag:
          predicate = Operands->car;
          body      = Operands->cdr;
          tag = Tag_of(body);
          switch (tag)
            { case NUL_tag:
                return Main_Error_Text(TFX_error_string,
                                       Main_While_String);
              case PAI_tag:
                compiled_predicate = compile_expression(predicate);
                compiled_body = compile_sequence(body,
                                                 1);
                compiled_while = make_WHI(compiled_predicate,
                                          compiled_body);
                return compiled_while; }
          return Main_Error_Text(ITF_error_string,
                                 Main_While_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_While_String); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ form ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_form(PAI_type Form)
  { EXP_type operands,
             operator;
    operator = Form->car;
    operands = Form->cdr;
    if (operator == Main_Begin)
          return compile_begin(operands);
    if (operator == Main_Define)
         return compile_define(operands);
    if (operator == Main_If)
             return compile_if(operands);
    if (operator == Main_Lambda)
         return compile_lambda(operands);
    if (operator == Main_Quote)
          return compile_quote(operands);
    if (operator == Main_Set)
            return compile_set(operands);
    if (operator == Main_While)
          return compile_while(operands);
    return compile_application(operator,
                               operands); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ expression ----------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_expression(EXP_type Expression)
  { TAG_type tag;
    tag = Tag_of(Expression);
    switch (tag)
      { case PAI_tag:
            return compile_form(Expression);
        case SYM_tag:
          return compile_symbol(Expression);
        case CHA_tag:
        case FLS_tag:
        case NBR_tag:
        case NUL_tag:
        case REA_tag:
        case STR_tag:
        case TRU_tag:
        case VEC_tag:
           return compile_value(Expression);
        case APL_tag:
        case BEG_tag:
        case CFN_tag:
        case CNT_tag:
        case DFF_tag:
        case DFV_tag:
        case DFZ_tag:
        case ENV_tag:
        case IFF_tag:
        case IFZ_tag:
        case LMB_tag:
        case LMZ_tag:
        case NAT_tag:
        case PRC_tag:
        case PRZ_tag:
        case QUO_tag:
        case USP_tag:
        case WHI_tag:
          Main_Error_Tag(ILT_error_string,
                         tag); }
    return Main_Error_Tag(IXT_error_string,
                          tag); }

/*------------------------------ public functions ----------------------------*/

EXP_type Compile_Compile(EXP_type Expression)
  { return compile_expression(Expression); }

NIL_type Compile_Initialize(NIL_type)
  { return; }
