                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*  version 7: iterative constructs  */
                     /*-----------------------------------*/
                     /*              reader               */
                     /*-----------------------------------*/

#include <stdlib.h>
#include <string.h>

#include "SlipMain.h"

#include "SlipPool.h"
#include "SlipRead.h"
#include "SlipScan.h"

/*------------------------------ private prototypes --------------------------*/

static EXP_type read_expression(NIL_type);

/*------------------------------ private constants ---------------------------*/

static const TXT_type IRB_error_string = "illegal right bracket";
static const TXT_type IRP_error_string = "illegal right parenthesis";
static const TXT_type IUP_error_string = "illegal use of period";
static const TXT_type PEP_error_string = "premature end of program";
static const TXT_type XCT_error_string = "excess tokens";

/*------------------------------ private variables ---------------------------*/

static SCA_type Token;

/*------------------------------ private functions ---------------------------*/

static NIL_type next_token(NIL_type)
  { Token = Scan_Next(); }

static EXP_type next_token_and_return(EXP_type Exp)
  { next_token();
    return Exp; }

static NIL_type build_list(PAI_type Pair)
  { EXP_type expression;
    PAI_type pair;
    switch (Token)
      { case RPR_token:
          return;
        case PER_token:
          next_token();
          expression = read_expression();
          Pair->cdr = expression;
          return; }
    expression = read_expression();
    pair = make_PAI(expression,
                    Main_Null);
    Pair->cdr = pair;
    build_list(pair); }

static VEC_type build_vector(UNS_type Count)
  { EXP_type expression;
    VEC_type vector;
    expression = read_expression();
    if (Token == RBR_token)
      vector = make_VEC(Count);
    else
      vector = build_vector(Count + 1);
    vector[Count] = expression;
    return vector; }

/*----------------------------------------------------------------------------*/

static EXP_type read_bracket(NIL_type)
  { Main_Error(IRB_error_string);
    return Main_Null; }

static CHA_type read_character(NIL_type)
  { CHA_type character;
    character = make_CHA(Scan_String[0]);
    return next_token_and_return(character); }

static EXP_type read_end_program(NIL_type)
  { Main_Error(PEP_error_string);
	  return Main_Null; }

static FLS_type read_false(NIL_type)
  { return next_token_and_return(Main_False); }

static PAI_type read_list(NIL_type)
  { EXP_type expression;
    PAI_type list;
    next_token();
    if (Token == RPR_token)
      list = Main_Empty_Pair;
    else
      { expression = read_expression();
        list = make_PAI(expression,
                        Main_Null);
        build_list(list); }
    return next_token_and_return(list); }

static NBR_type read_number(NIL_type)
  { LNG_type raw_number;
    NBR_type number;
    raw_number = atoll(Scan_String);
    number = make_NBR(raw_number);
		return next_token_and_return(number); }

static EXP_type read_parenthesis(NIL_type)
  { Main_Error(IRP_error_string);
    return Main_Null; }

static REA_type read_real(NIL_type)
  { FLO_type raw_real;
    REA_type real;
    raw_real = atof(Scan_String);
    real = make_REA(raw_real);
    return next_token_and_return(real); }

static EXP_type read_period(NIL_type)
  { Main_Error(IUP_error_string);
	  return Main_Null; }

static EXP_type read_quote(NIL_type)
  { EXP_type expression;
    PAI_type pair;
    next_token();
		expression = read_expression();
    pair = make_PAI(expression,
                    Main_Null);
    pair = make_PAI(Main_Quote,
                    pair);
    return pair; }

static STR_type read_string(NIL_type)
  { STR_type string;
    string = make_STR(Scan_String);
    return next_token_and_return(string); }

static SYM_type read_symbol(NIL_type)
  { SYM_type symbol;
    symbol = Pool_Enter(Scan_String);
    return next_token_and_return(symbol); }

static TRU_type read_true(NIL_type)
  { return next_token_and_return(Main_True); }

static VEC_type read_vector(NIL_type)
  { VEC_type vector;
    next_token();
    if (Token == RBR_token)
      vector = Main_Empty_Vector;
    else
      vector = build_vector(1);
    return next_token_and_return(vector); }

static EXP_type read_expression(NIL_type)
  { switch (Token)
      { case CHA_token:
          return read_character();
        case END_token:
          return read_end_program();
        case FLS_token:
          return read_false();
        case IDT_token:
          return read_symbol();
        case LBR_token:
          return read_vector();
        case LPR_token:
          return read_list();
        case NBR_token:
          return read_number();
        case PER_token:
          return read_period();
        case QUO_token:
          return read_quote();
        case RBR_token:
          return read_bracket();
        case RPR_token:
          return read_parenthesis();
        case REA_token:
          return read_real();
        case STR_token:
          return read_string();
        case TRU_token:
          return read_true(); }
	  return Main_Null; }

/*----------------------------- public functions -----------------------------*/

NIL_type Read_Initialize(NIL_type)
  { Scan_Initialize(); }

EXP_type Read_Parse(TXT_type Input)
  { EXP_type expression;
    Scan_Preset(Input);
    next_token();
    expression = read_expression();
    if (Token != END_token)
      Main_Error(XCT_error_string);
    return expression; }
